#! /usr/bin/env python
# -*- mode:python -*-
# $Id: keeper_example.py 11527 2006-04-13 00:23:12Z vax $

"""
Sample script for using dfd_keeper.

TODO: Add knowledge of internal/external IPs and modify rules to check errors.
"""

# This is where you set the TCP port to listen on.
listen_port = 8007

# Parse command-line options.
import getopt, sys

import textwrap

from dfd_keeper import *

state_file="keeper_example.state"
nostate=False
verbose=True

usage = "Usage: %s [--help] [--syntax] [--test] [--port NNNN] [--wan ifc] [--lan ifc]" \
        % sys.argv[0]
try:
    opts, args = getopt.getopt(sys.argv[1:], "hstp:w:l:",
                               [ "help", "syntax", "test", "port=", "wan=", "lan=" ])
except getopt.GetoptError:
    raise SystemExit, usage
mode = "normal"
# NOTE: These are right for me, but probably not you.
d = { "wan_if" : magic("xl2"), "lan_if" : magic("xl1"), "bt_client" : magic(), "block" : magic() }
for (o, a) in opts:
    if o in ("-h", "--help"):
        print usage
        raise SystemExit
    if o in ("-s", "--syntax"):
        mode = "syntax"
    if o in ("-t", "--test"):
        # Don't actually write any changes to the firewall.
        mode = "test"
    if o in ("-w", "--wan"):
        d["wan_if"] = a
    if o in ("-l", "--lan"):
        d["lan_if"] = a
    if o in ("-p", "--port"):
        listen_port = int(a)

s = dict()
if not nostate:
    try:
        if state_file == None:
            if verbose: print "Setting default state file"
            s = set_persist_hooks()
        else:
            if verbose: print "Setting state file %s" % state_file
            s = set_persist_hooks(state_file)
    except FileNotFoundError:
        if verbose: print "No state file created."
        nostate = True

# See if any variables have been added or deleted.
# If they haven't, use the old state.
# NOTE: If nostate is set, then s will be empty and
# so this will not hurt anything.
if s.has_key("namespace"):
    skeys = s["namespace"].keys()
    skeys.sort()
    dkeys = d.keys()
    dkeys.sort()
    if skeys == dkeys:
        if verbose: print "No changed variables, using old namespace."
        d = s["namespace"]
else:
    if verbose: print "Using new namespace."

# No saved packet_filter object, create one.
packet_filter = pf(namespace=d, mode=mode)
nat = packet_filter["translation"]
# NOTE: This assumes your WAN interface could be DHCP.
# If it isn't, you don't need the parens around $wan_if.
nat.make_rule("nat on $wan_if from $lan_if:network to any -> ($wan_if)")
# As an example of how to do p2p behind NAT, set up the ability to
# port forward back to a client.  See the dfd_sniff program for how to
# automatically trigger this rule.
nat.make_rule("""
rdr pass on $wan_if proto { udp tcp } from any to ($wan_if) port { 6881:6890 } -> $bt_client
""")
# Retrieve the filter table.
f = packet_filter["filter"]
f.make_rule("""
# Allow all loopback traffic.
pass quick on lo0

# Default deny.
block all

# Allow LAN to bomb out quickly.
block return on $lan_if

# Don't allow other networks to impersonate LAN.
antispoof quick for $lan_if

# Block leakage of LAN stuff to anywhere else.
block out log quick on ! $lan_if to $lan_if:network
""")
f.make_rule("""
# Block hosts we have specified in both directions.
block in log quick on $wan_if from $block to any
block out log quick on $wan_if from any to $block
""")
f.make_rule("""
# Allow firewall to talk to LAN.
pass out quick on $lan_if from $lan_if to $lan_if:network keep state

# Allow anything in from LAN that isn't destined to the LAN.
pass in quick on $lan_if to ! $lan_if:network keep state allow-opts

# Allow LAN hosts to SSH into this box.
pass in quick on $lan_if proto tcp from any to $lan_if port ssh flags S/SA
""")
f.make_rule("""
# Allow connections out WAN, and randomize SEQ #s.
pass out quick on $wan_if all modulate state allow-opts
""", tag="wan")

if not s.has_key("packet_filter") or repr(packet_filter) != repr(s["packet_filter"]):
        s["packet_filter"] = packet_filter
        s["namespace"] = d

        # Flush all states.
        packet_filter.flush()

        # Sync once at startup.
        packet_filter.force_sync()

# This code illustrates how you might use this module in a program.
# It listens to a port and accepts CRLF-terminated commands.
# Actually the CR is optional, for compatibility with netcat.

class command_dispatch(helper_functions):
    """
    This class defines a command suite for pf.

    When initializing it, you pass it a configured set of rules.
    By defining commands of the form cmd_whatever, you make that
    command available to anyone to use via tcp_command_taker.

    Implementor note: to print to the client, just return a string!
    """

    # This is what all the user-accessible commands begin with.
    prefix="cmd_"

    def __init__(self, packet_filter):
        self.pf = packet_filter

        # Create the block list command.
        block_help = """\
        block [add|del] host
        Block an IP from sending in data via WAN interface either direction.
        XXX: Assumes it is on the remote side of that interface.
        """
        self.list_factory("block", d["block"],"add",
                          textwrap.dedent(block_help).strip())

        # Here is how you specify a custom help message for the command.
        help="Switches on/off connectivity with the Internet.\n" + \
              "For emergencies only!\n"
        self.switch_factory("wan", help=help, flushing=True)
        # In addition to toggling active flags on wan-tagged rules,
        # you could also just set d["wan_if"] = "" and that would
        # disable any rule with $wan_if in it.

        # Pull in stuff from helper_functions.
        for c in (helper_functions.end_user):
            setattr(self.__class__, self.prefix + c,
                    getattr(helper_functions, c))

        # Sync once at startup.
        packet_filter.force_sync()

    # The rest of the user-accessible commands follow.

    def cmd_drop_state(self, src, dst=None):
        """Drop a particular state table entry.  Takes src and optional dst."""
        src = str(ipv4.address(src))
        if dst == None:
            self.pf.flush(src)
        else:
            self.pf.flush(src, str(ipv4.address(dst)))

    def cmd_bittorrent(self, host=""):
        """Specify the bittorrent client, or nothing to turn off forwarding."""
        t = self.pf.namespace["bt_client"]
        if t != "": self.pf.flush(t)
        self.pf.namespace["bt_client"] = magic(host)

def command_factory(p=s["packet_filter"]):

    # Create a command module to use for executing commands.
    command_module = command_dispatch(p)

    # Set this up so that it syncs the rules every time you enter a command.
    synch_cmd = sync_proxy(command_module, command_module._sync)

    return synch_cmd

# Listen to a TCP port for commands.
# NOTE: This binds to the localhost interface only.
# Do not change this unless you really know what you are doing.  If you
# know what you're doing, remove the last argument and it will bind to
# the wildcard address.

reactor.listenTCP(listen_port, tcp_factory(command_factory, tcp_command_taker),
                  interface='localhost')

# Start running.
reactor.run()
